use base64::Engine;
use log::{debug, info};
use quick_xml::{
    events::{attributes::Attributes, BytesDecl, BytesEnd, BytesStart, Event},
    Writer,
};
use tokio::io::AsyncWrite;

use crate::{error::StreamError, tag};

type Result<T> = std::result::Result<T, StreamError>;

pub async fn error<W: AsyncWrite + Unpin>(writer: W, err: StreamError) -> Result<()> {
    let mut writer = Writer::new(writer);
    let err = err.to_string();
    writer
        .write_event_async(Event::Start(BytesStart::new(tag::ERROR_ELEMENT)))
        .await?;
    writer
        .write_event_async(Event::Start(
            BytesStart::new(&err)
                .with_attributes(vec![("xmlns", "urn:ietf:params:xml:ns:xmpp-streams")]),
        ))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(&err)))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(tag::ERROR_ELEMENT)))
        .await?;
    Ok(())
}

pub async fn write_stream_header<W: AsyncWrite + Unpin>(
    writer: W,
    req: StreamHeader,
) -> Result<()> {
    let mut writer = Writer::new(writer);
    writer
        .write_event_async(Event::Decl(BytesDecl::new("1.0", Some("utf-8"), None)))
        .await?;
    let id = req.id.unwrap_or_default();
    let (from, to) = (req.from.unwrap_or_default(), req.to.unwrap_or_default());

    let mut attrs = vec![
        ("id", id.as_str()),
        ("xmlns:stream", "http://etherx.jabber.org/streams"),
        ("xmlns", "jabber:client"),
        ("xml:lang", "en"),
        ("version", "1.0"),
    ];
    if !to.is_empty() {
        attrs.push(("to", to.as_str()));
    }
    if !from.is_empty() {
        attrs.push(("from", from.as_str()));
    }
    writer
        .write_event_async(Event::Start(
            BytesStart::new("stream:stream").with_attributes(attrs),
        ))
        .await?;

    Ok(())
}

#[derive(Debug, Clone)]
pub struct StreamHeader {
    pub from: Option<String>,
    pub to: Option<String>,
    pub id: Option<String>,
    pub namespace: XMLNamespace,
}

impl TryFrom<Attributes<'_>> for StreamHeader {
    type Error = StreamError;

    fn try_from(value: Attributes<'_>) -> std::result::Result<Self, Self::Error> {
        let mut from: Option<String> = None;
        let mut to: Option<String> = None;
        let mut ns: Option<XMLNamespace> = None;
        for v in value {
            let v = v?;
            match v.key.local_name().into_inner() {
                b"from" => {
                    from = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"to" => {
                    to = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"xmlns" => match v.value.to_vec().as_slice() {
                    b"jabber:client" => {
                        ns = Some(XMLNamespace::JabberClient);
                    }
                    _ => return Err(StreamError::InvalidNamespace),
                },
                other => {
                    info!(
                        "ignoring key {} with value << {} >>",
                        String::from_utf8(other.to_vec()).unwrap_or_default(),
                        String::from_utf8(v.value.to_vec()).unwrap_or_default(),
                    );
                }
            }
        }
        debug!("from: {from:?}; to: {to:?}");
        Ok(StreamHeader {
            from: from,
            to: Some(to.ok_or(StreamError::HostUnknown)?),
            id: None,
            namespace: ns.ok_or(StreamError::BadNamespacePrefix)?,
        })
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum XMLNamespace {
    JabberClient,
}

pub fn gen_id() -> String {
    let buf: [u8; 16] = rand::random();
    let mut output = String::new();
    base64::engine::general_purpose::STANDARD_NO_PAD.encode_string(buf, &mut output);

    output
}
