use std::sync::Arc;

use async_trait::async_trait;
use quick_xml::{
    events::{BytesStart, Event},
    Writer,
};
use tokio::{io::AsyncWrite, net::TcpStream};
use tokio_rustls::{rustls, server::TlsStream, TlsAcceptor};

use crate::{
    error::StreamError,
    tag::{self, Tag},
};

pub enum Step {
    Proceed,
    Failure,
}

#[async_trait]
impl Tag for Step {
    async fn write_tag<W>(&self, writer: W) -> Result<(), crate::error::StreamError>
    where
        W: AsyncWrite + Unpin + Send,
    {
        let mut writer = Writer::new(writer);
        writer
            .write_event_async(Event::Empty(
                BytesStart::new(match self {
                    Step::Proceed => "proceed",
                    Step::Failure => "failure",
                })
                .with_attributes(vec![("xmlns", tag::TLS_NAMESPACE)]),
            ))
            .await?;
        Ok(())
    }
}

pub(crate) enum TcpConnOrTLS {
    TCP(TcpStream, StreamError),
    /// Failed is returned when there's a failure starting the TLS stream
    /// thus the TCP stream is consumed.
    Failed(StreamError),
    TLS(TlsStream<TcpStream>),
}

pub(crate) async fn start_tls(
    conn: TcpStream,
    start_tls_event: BytesStart<'_>,
    tls_config: Arc<rustls::ServerConfig>,
) -> TcpConnOrTLS {
    let mut conn = conn;
    match start_tls_event.try_get_attribute("xmlns") {
        Ok(namespace) => {
            if &namespace
                .map(|a| String::from_utf8(a.value.as_ref().to_vec()).unwrap_or_default())
                .unwrap_or_default()
                != tag::TLS_NAMESPACE
            {
                if let Err(err) = Step::Failure.write_tag(&mut conn).await {
                    return TcpConnOrTLS::TCP(conn, err);
                }
                return TcpConnOrTLS::TCP(conn, StreamError::InvalidNamespace);
            }
        }
        Err(_) => {
            if let Err(err) = Step::Failure.write_tag(&mut conn).await {
                return TcpConnOrTLS::TCP(conn, err);
            }
            return TcpConnOrTLS::TCP(conn, StreamError::BadFormat);
        }
    }

    if let Err(err) = Step::Proceed.write_tag(&mut conn).await {
        return TcpConnOrTLS::TCP(conn, err);
    };
    let acceptor = TlsAcceptor::from(tls_config);
    match acceptor.accept(conn).await {
        Ok(tls_conn) => TcpConnOrTLS::TLS(tls_conn),
        Err(err) => TcpConnOrTLS::Failed(err.into()),
    }
}
