use std::sync::Arc;

use log::{error, info};
use tokio::net::TcpListener;
use tokio_rustls::{rustls, TlsAcceptor};

use crate::{streamstart, tls::stream};

pub async fn listen_starttls(
    hostname: String,
    insecure_port: u16,
    tls_config: Arc<rustls::ServerConfig>,
) -> Result<(), anyhow::Error> {
    let listener = TcpListener::bind(("0.0.0.0", insecure_port)).await?;
    loop {
        match listener.accept().await {
            Ok(conn) => {
                info!("opening starttls connection from {}", conn.1);
                streamstart::spawn(hostname.clone(), conn, tls_config.clone());
            }
            Err(e) => {
                eprintln!("listening: {e}");
                continue;
            }
        }
    }
}

pub async fn listen_tls(
    hostname: String,
    tls_port: u16,
    tls_config: Arc<rustls::ServerConfig>,
) -> Result<(), anyhow::Error> {
    let listener = TcpListener::bind(("0.0.0.0", tls_port)).await?;
    let acceptor = TlsAcceptor::from(tls_config);
    loop {
        match listener.accept().await {
            Ok(conn) => {
                info!("opening TLS connection from {}", conn.1);
                match acceptor.accept(conn.0).await {
                    Ok(conn) => {
                        let cloned_hostname = hostname.clone();
                        tokio::spawn(async move {
                            stream::TLSStream::new(conn, cloned_hostname)
                                .start_stream()
                                .await
                        });
                    }
                    Err(err) => error!("TLS accept error for {}: {err}", conn.1),
                }
            }
            Err(e) => {
                eprintln!("listening: {e}");
                continue;
            }
        }
    }
}
